/*
** Module   : AdminLogon.java
** Abstract : Creates logon panel and manages user's logon/logoff.
**
** Copyright (c) 2009-2011, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description----------------------
** 001 NVS 20090116   @41619 Created the first version of the file.
** 002 NVS 20090318   @41631 Resetting cache serial number at logoff.
** 003 NVS 20090323   @41665 Refactored one more time to get rid of static
**                           members.
** 004 NVS 20090326   @41689 Logging off switches to the blank screen. 
** 005 NVS 20090428   @42038 Reworked event notifications.
** 006 NVS 20090429   @42080 Redone bracketing of server calls with busy/idle
**                           mouse cursors.
** 007 CA  20090529   @42489 Added getter for the net session.
** 008 CA  20090604   @42565 Added a logoff version which doesn't display the
**                           confirmation dialog.
** 009 CA  20090610   @42654 If session is terminated abru9tly (not using 
**                           logoff) the user will be logged out and notified.
** 010 GES 20090623   @42780 Cleaned up error processing and messages.
** 011 NVS 20090807   @43549 Refactored in support of custom menu extensions.
** 012 NVS 20090816   @43649 Added static autoLogoff() method.
** 013 NVS 20090909   @43852 Logoff is issued if the account doesn't have
**                           sufficient admin rights.
** 014 NVS 20090930   @44102 Detailed diagnostics of all errors related to the
**                           loading of resource editors.
** 015 NVS 20091007   @44112 Removed some debugging related modifications.
** 016 SIY 20091216   @44513 Added checking for pending refresh at logoff.
** 017 SVL 20100211   @44575 Added getLoggedInUser() function.
** 018 SVL 20110324          Enable the secure mode and set the secure port
**                           into bootstrap config.
*/

package com.goldencode.p2j.admin.client;

import com.goldencode.p2j.admin.AdminExports;
import com.goldencode.p2j.admin.AdminDef;
import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.util.*;

import java.awt.*;
import java.awt.event.*;
import java.lang.reflect.*;
import java.util.Map;
import java.util.HashMap;
import javax.swing.*;

/**
 * Logon screen management class. Creates the logon screen, handles the
 * GUI input and creates a session with the P2J admin server.
 * Logs the user off upon request. 
 *
 * @author NVS
 */
public class AdminLogon
implements SessionListener
{
   /** Instance of this */
   private static AdminLogon logon = null;
   
   /** Related instance of AdminClient */
   private AdminClient client = null;

   /** Related instance of AdminMenu */
   private AdminMenu menu = null;

   /** Related instance of AdminTarget */
   private AdminTarget target = null;

   /** instance of the logon panel */
   private JPanel outerPanel = null;
   
   /** instance of the real logon panel */
   private JPanel panel = null;
   
   /** 
    * BootstrapConfig used in applet.
    * This reference is to an instance created in AdminClient.
    */
   private BootstrapConfig bc = null;

   /** references the host string in AdminClient */
   private String host = null;
   
   /** references the port string in AdminClient */
   private String port = null;

   /** instance of session with the admin server */
   private Session session = null;

   /** instance of the AdminExports interface */
   private AdminExports adm = null;

   /** PasswordField used in in logon panel */
   private JPasswordField pwd = null;

   /** UserID TextField used in logon panel */
   private JTextField uid = null;

   /** the instance of the account extension plugin */
   private ClientAccountExtension accountExt = null;
   
   /** the instance of the menu extension plugin */
   private AdminCustomMenu menuExt = null;
   
   /** the custom ACL editors. */
   private Map<String, RightsEditor> customAclEditors = null;

   /** resource types */
   private String[] resourceTypes = null;
   
   /**
    * Creates logon screen populated with the usual things like user ID and
    * password fields with labels and the logon button.
    * 
    * @param    bc
    *           reference to the instance of BootstrapConfig
    * 
    * @param    host
    *           admin server host name to connect to
    * 
    * @param    port
    *           admin server port to connect to
    */
   AdminLogon(BootstrapConfig bc, String host, String port) 
   {
      logon = this;
      
      // remember the parameters
      this.bc = bc;
      this.host = host;
      this.port = port;
      
      // logon panel
      outerPanel = new JPanel();
      
      // centered smaller real panel
      panel = new JPanel();
      
      // and this is the trick to get the small panel centered:
      // vertical box layout with two glues around a nested
      // horizontal box layout with two glues around the small panel
      outerPanel.setLayout(new BoxLayout(outerPanel, BoxLayout.PAGE_AXIS));
      outerPanel.add(Box.createVerticalGlue());
      JPanel innerPanel = new JPanel();
      outerPanel.add(innerPanel);
      outerPanel.add(Box.createVerticalGlue());
      innerPanel.setLayout(new BoxLayout(innerPanel, BoxLayout.LINE_AXIS));
      innerPanel.add(Box.createHorizontalGlue());
      innerPanel.add(panel);
      innerPanel.add(Box.createHorizontalGlue());
      
      // the small real panel uses group layout
      GroupLayout layout = new GroupLayout(panel);
      panel.setLayout(layout);
      layout.setAutoCreateGaps(true);
      layout.setAutoCreateContainerGaps(true);
      
      // components
      JLabel luid = new JLabel("User ID");
      JLabel lpwd = new JLabel("Password");
      uid = new JTextField(16);
      pwd = new JPasswordField(16);
      
      // Create a sequential group for the horizontal axis:
      {
         GroupLayout.SequentialGroup hGroup = 
            layout.createSequentialGroup();
         
         // parallel group of labels
         hGroup.addGroup(
            layout.createParallelGroup().
            // id label
            addComponent(luid).
            // pw label
            addComponent(lpwd)
         );
         
         // parallel group of input fields
         hGroup.addGroup(
            layout.createParallelGroup().
            // uid field
            addComponent(uid).
            // pwd field
            addComponent(pwd)
         );
         
         layout.setHorizontalGroup(hGroup);
      }
      
      // Create a sequential group for the vertical axis
      {
         GroupLayout.SequentialGroup vGroup = layout.createSequentialGroup();
         vGroup.
            // parallel group of luid, uid
            addGroup(layout.createParallelGroup(
                                              GroupLayout.Alignment.BASELINE).
               // luid
               addComponent(luid).
               // uid
               addComponent(uid)
               ).
            // parallel group of lpwd, pwd
            addGroup(layout.createParallelGroup(
                                              GroupLayout.Alignment.BASELINE).
               // lpwd
               addComponent(lpwd).
               // pwd
               addComponent(pwd)
               )
            ;
            
         layout.setVerticalGroup(vGroup);
      }
      
      layout.linkSize(luid, lpwd);
      layout.linkSize(uid, pwd);
      
      // clear inputs
      uid.setText("");
      pwd.setText("");
      
      // attach event listeners
      uid.addActionListener(
         new ActionListener()
         // this class reacts to user id field completion
         {
            @SuppressWarnings("unused") 
            public void actionPerformed(ActionEvent evt) 
            {
               completeUid();
            }
         });

      pwd.addActionListener(
         new ActionListener() 
         // this class reacts to password field completion
         {
            @SuppressWarnings("unused") 
            public void actionPerformed(ActionEvent evt) 
            {
               logon();
            }
         });
   }

   /**
    * Handle abrupt session termination. This displays an error message and
    * logs off.
    * 
    * @param    session
    */
   public void terminate(Session session)
   {
      AdminUtils.errorDialog(client,
                             "Connection Lost", 
                             "The session was abruptly terminated. Logging " +
                             "off.");

      // destroy the session and exports, as connection is already dead
      session = null;
      adm     = null;

      logoff(false);
   }

   @Override
   public void initialize(Session session)
   {
      // TODO: implement it
   }


   /**
    * Get the current P2J session.
    * 
    * @return  current P2J session.
    */
   Session getSession()
   {
      return session;
   }

   /**
    * Initializes useful references.
    * 
    * @param    client
    *           instance of AdminClient
    * 
    * @param    menu
    *           instance of AdminMenu
    * 
    * @param    target
    *           instance of AdminTarget
    */
   void init(AdminClient client, AdminMenu menu, AdminTarget target)
   {
      this.client = client;
      this.menu = menu;
      this.target = target;
   }

   /**
    * Returns the instance of the logon screen.
    * 
    * @return   instance of the logon screen to be added to the screen swapper
    */
   JPanel getLogonPanel()
   {
      return outerPanel;
   }
   
   /**
    * Switches to the logon panel allowing user log on.
    */
   void logonPanel()
   {
      // bring the logon screen
      client.show(AdminClient.SCREEN_LOGON,
                  "Please log on @" + client.getTargetServer());

      // set initial focus on the user id field
      uid.requestFocusInWindow();
   }

   /**
    * Returns the server exports.
    * 
    * @return   the instance of the remote server admin interface
    */
   AdminExports getAdminExports()
   {
      return adm;
   }
   
   /**
    * This method is called when connect button is clicked It creates and
    * caches a session with the P2J admin server.
    */
   private void logon() 
   {
      // get the user id from the form and verify it
      String id = uid.getText().trim();
      
      if (id.length() == 0)
      {
         // set initial focus on the user id field
         uid.requestFocusInWindow();
         
         // alarm
         client.beep("type user ID");
         return;
      }
      
      // get the password from the form - no verification here
      String pw = new String(pwd.getPassword());
      
      // stuff the bootstrap config
      bc.setConfigItem("net", "queue", "conversation", "true");
      bc.setConfigItem("net", "queue", "start_thread", "false");
      bc.setConfigItem("net", "server", "host", host);
      bc.setConfigItem("net", "server", "port", port);
      bc.setConfigItem("security", "certificate", "validate", "false");
      bc.setConfigItem("security", "authentication", "type", "program");
      bc.setConfigItem("access", "subject", "id", id);
      bc.setConfigItem("access", "password", "user", pw);
      bc.setConfigItem("net", "connection", "secure", "true");
      bc.setConfigItem("net", "server", "secure_port", port);

      // attempt session creation
      try
      {
         // busy cursor
         AdminUtils.busy(client);
         
         // disable inputs temporarily
         uid.setEnabled(false);
         pwd.setEnabled(false);
         
         // instantiate network session manager
         SessionManager sessMgr = SessionManagerFactory.createLeafNode(bc);
                           
         // establish a session with the P2J server
         session = sessMgr.connectDirect(new InterruptedExceptionHandler(),
                                         null);
         
         session.addSessionListener(this);
      }
      catch (Exception e)
      {
         // set initial focus on the user id field
         uid.setEnabled(true);
         pwd.setEnabled(true);
         uid.requestFocusInWindow();
         
         String txt = e.getMessage();
         String err = String.format("Logon failed (reason: %s). Retry is " +
                                    "allowed.",
                                    (txt == null) ? e.toString() : txt); 
         
         // alarm
         client.beep(err);
         
         // idle cursor
         AdminUtils.idle(client);
         return;
      }
      finally
      {
         uid.setEnabled(true);
         pwd.setEnabled(true);
         
         // idle cursor
         AdminUtils.idle(client);
      }
      
      // clear inputs
      uid.setText("");
      pwd.setText("");
      
      // reflect new state
      menu.connectedState();
      client.show(AdminClient.SCREEN_BLANK, "");

      // get the admins server APIs
      try
      {
         adm = (AdminExports)RemoteObject.obtainNetworkInstance(
            AdminExports.class, session);
      }
      catch (RuntimeException rte)
      {
         // no rights to access the interface - bail out
         logoff(false);
         
         // show fatal dialog
         AdminUtils.errorDialog(client, "Fatal error", 
               "No rights to access the admin interface. Logging off.");

         // idle cursor
         AdminUtils.idle(client);
         return;
      }
      
      // check the admin rights and get the server name
      String serverName = null;
      
      try
      {
         serverName = adm.getServerName();
      }
      catch (RuntimeException rte)
      {
         // no rights to access the interface - bail out
         logoff(false);
         
         // show fatal dialog
         AdminUtils.errorDialog(client, "Fatal error", 
               "No rights to access the admin interface. Logging off.");

         // idle cursor
         AdminUtils.idle(client);
         return;
      }
      
      if (serverName == null)
      {
         // no admin rights - bail out
         logoff(false);
         
         // show fatal dialog
         AdminUtils.errorDialog(client, "Fatal error", 
               "No adminisitration rights given. Logging off.");

         // idle cursor
         AdminUtils.idle(client);
         return;
      }
      
      // set the target description
      client.setTargetInfo(AdminClient.TARGET_SERVER, serverName);

      // notify other classes 
      client.initExports(adm);
      target.initExports(adm);
      
      // choose target --> live automatically
      menu.getTargetLive().setSelected(true);
      target.goLive();
      
      // cache the admin definition values
      AdminDef adminDef = adm.getAdminDef();

      if (!loadCustomExtensionsPlugin(adminDef.customClientExt))
      {
         logoff(false);
         // show fatal dialog
         AdminUtils.errorDialog(client, "Fatal error", 
               "Error loading admin client extension plugin. Logging off.");

         // idle cursor
         AdminUtils.idle(client);
         return;
      }

      if (!loadCustomAclEditors(adminDef))
      {
         logoff(false);
         // show fatal dialog
         AdminUtils.errorDialog(client, "Fatal error", 
               "Error loading custom ACL editors. Logging off.");

         // idle cursor
         AdminUtils.idle(client);
         return;
      }

      // post logon init
      client.postLogon(adminDef, resourceTypes, accountExt, menuExt, 
                       customAclEditors);
      
      // idle cursor
      AdminUtils.idle(client);
   }
   
   /**
    * Logs the user off the admin server, with a confirmation dialog before
    * actual logoff.
    */
   void logoff()
   {
      logoff(true);
   }

   /**
    * Logs the user off the admin server, with a confirmation dialog before
    * actual logoff.
    */
   static void autoLogoff()
   {
      if (logon != null)
      {
         logon.logoff(false);
      }
   }

   /**
    * Get the name of the user currently logged into admin console.
    *
    * @return The name of the user currently logged into admin console or
    *         <code>null</code> if there is no logged in user. 
    */
   public static String getLoggedInUser()
   {
      String user = null;
      if (logon != null && logon.getSession() != null)
      {
         try
         {
            user = logon.bc.getConfigItem("access", "subject", "id");
         }
         catch (ConfigurationException e)
         {
            // suppress
         }
      }

      return user;
   }

   /**
    * Logs the user off the admin server.
    * 
    * @param    confirm
    *           <code>true</code> if confirmation dialog should be displayed.
    */
   void logoff(boolean confirm)
   {
      if (confirm &&
          !AdminUtils.
          confirmDialog(client, "Logoff Confirmation",
                        "You are about to be logged off. Continue?"))
      {
         // declined
         return;
      }
      
      if (confirm && 
          adm.isRefreshPending() &&
          AdminUtils.
          confirmDialog(client, "Refresh Confirmation",
                        "Warning: there are changes to the directory which " +
                        "have not yet been re-read into the security cache.\n" +
                        "Refresh cache now?"))
      {
         target.refresh();
      }
      
      // shut down session
      if (session != null)
      {
         // explicit termination - remove the listener
         session.removeSessionListener(this);
         session.terminate();
         session = null;
         adm = null;
      }

      // clear aged data
      accountExt = null;
      menuExt = null;
      customAclEditors = null;
      resourceTypes = null;
      
      // signal logoff state
      client.notifyEvent(AdminNotify.NOTIFY_LOGOFF);
      
      // switch the screen back to logon panel
      logonPanel();
   }
   
   /**
    * Reacts to ENTER in the user ID field by focusing the password field.
    */
   private void completeUid() 
   {
      // get the user id from the form and verify it
      String id = uid.getText().trim();
      
      if (id.length() == 0)
      {
         // set initial focus on the user id field
         uid.requestFocusInWindow();
         
         // alarm
         Toolkit.getDefaultToolkit().beep();
         return;
      }

      // transfer focus to the password field
      pwd.requestFocusInWindow();
   }
   
   /**
    * Load and instantiate the custom client extensions plugin. It's
    * responsible for all account extensions and the custom menu items.
    * 
    * @return   <code>true</code> if loading succeeded
    */
   private boolean loadCustomExtensionsPlugin(String pluginClass)
   {
      if (pluginClass != null && pluginClass.length() > 0)
      {
         try
         {
            Class<?> clazz = Class.forName(pluginClass);
            Object inst = clazz.newInstance();
            menuExt = (AdminCustomMenu)inst;
            accountExt = (ClientAccountExtension)inst;
            menuExt.attach(client);
         }
         catch (Exception e)
         {
            client.beep("could not load plugin class " + pluginClass);
            return false;
         }
      }

      return true;
   }
   
   /**
    * Load and instantiate the custom ACL editors.
    * 
    * @return   <code>true</code> if loading succeeded
    */
   @SuppressWarnings("unchecked")
   private boolean loadCustomAclEditors(AdminDef adminDef)
   {
      customAclEditors = new HashMap<String, RightsEditor>();
      resourceTypes = new String[adminDef.customAclEditors.size()];

      int i = 0;
      for (String key : adminDef.customAclEditors.keySet())
      {
         resourceTypes[i++] = key;
         
         String pluginClass = adminDef.customAclEditors.get(key);
         
         if (pluginClass != null && pluginClass.length() > 0)
         {
            Class<RightsEditor> clazz = null;
            
            try
            {
               clazz = (Class<RightsEditor>) Class.forName(pluginClass);
            }
            catch (ClassNotFoundException cnfe)
            {
               client.beep("plugin not found " + cnfe);
               return false;
            }
               
            Method m = null;
            
            try
            {
               m =clazz.getMethod("initialize", AdminClient.class, 
                                  Description[].class);
            }
            catch (NoSuchMethodException nsme)
            {
               client.beep("no initialize() method " + nsme);
               return false;
            }
            
            RightsEditor editor = null;

            try
            {
               editor = clazz.newInstance();
            }
            catch (InstantiationException ie)
            {
               client.beep("can't instantiate plugin " + ie);
               return false;
            }
            catch (IllegalAccessException iae)
            {
               client.beep("no access to plugin " + iae);
               return false;
            }
            
            try
            {
               m.invoke(editor, client, adminDef.resourceDescription.get(key));
            }
            catch (IllegalAccessException iae)
            {
               client.beep("no access to initialize in plugin " + iae);
               return false;
            }
            catch (InvocationTargetException ite)
            {
               Throwable t = ite.getCause();
               client.beep("initialize() failed " + t);
               return false;
            }
            
            customAclEditors.put(key, editor);
         }
         
      }
      
      return true;
   }
}
