/*
** Module   : Conversation.java
** Abstract : represents one side of a two-party single-threaded cooperative 
**            synchronous communication session
**
** Copyright (c) 2007-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- -------------------------- Description -----------------------------
** 001 GES 20070115   @31830 Created initial version which represents one
**                           side of a two-sided single-threaded 
**                           cooperative synchronous communication 
**                           session.
** 002 NVS 20070411   @32927 Conversation threads are transient and as
**                           such should call dropInitialSecurityContext()
**                           before they exit to maintain the balance of
**                           use for the initial security context.
** 003 ECF 20071101   @35878 Refactored net package. Use SessionManager to
**                           manage security contexts. Other minor
**                           cleanup.
** 004 ECF 20071127   @36050 Removed remaining SecurityManager reference.
**                           NetResource plugin is now retrieved from
**                           SessionManager. Ensured there is an initial
**                           security context set before getting plugin.
** 005 SVL 20080319   @37556 block() is interrupted if the queue is not
**                           running.
** 006 ECF 20090205   @42108 Fixed synchronization. The block() method
**                           could hold the instance monitor for an
**                           arbitrarily long time while business logic
**                           was executed, blocking a call to halt() on
**                           the reader thread from shutting down the
**                           transport after a reader socket error.
** 007 GES 20090722   @43332 On server-side, name the thread using context
**                           data.
** 008 GES 20090723   @43360 Reduced logging of normal session ends to FINE
**                           from INFO. Increased the usefulness of the entry
**                           when generated.
** 009 GES 20090816   @43636 Moved context processing into the try block to
**                           ensure that cleanup happens on failures.
** 010 OM  20131014          Added support for setting initially authenticated users in the 
**                           correct context.
*/

package com.goldencode.p2j.net;

import java.util.logging.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.util.LogHelper;

/**
 * This class represents one side of a two-sided single-threaded cooperative 
 * synchronous communication session.
 * <p>
 * This conversation mode concept has the attribute that when one side of the
 * session is active, the other side is blocked and vice versa.  This makes
 * all processing synchronous and single-threaded.  Control flow transfers
 * back and forth between each side of the session which is why this is
 * described as cooperative.
 * <p>
 * A dedicated thread exists on each side of such a session.  This thread
 * uses a re-entrant blocking function which uses the thread's stack to
 * naturally manages the call/return semantics of each reply to a synchronous
 * request.  This same blocking function is where all inbound synchronous
 * requests are processed (asynchronous requests are still processed in the
 * dispatcher).  
 * <p>
 * There are 2 ways of using this class: with a new daemon thread or
 * by capturing an already existing thread.  If you require a new thread
 * to be created for this processing, then use this class as a runnable
 * and then start the thread (set it as daemon mode first).  If you wish to
 * re-use an existing thread, then once the {@link Queue} shifts into
 * conversation mode, the next call into any synchronous message processing
 * will "capture" the thread for the duration of that processing (until the
 * reply message is received).  This is useful in cases where there is an
 * existing "main" thread which calls into a primary entry point and then
 * stays there for the duration of the session.  In this second (capture)
 * approach, it is important to note that the session will not properly 
 * respond to synchronous requests until the thread is captured.
 */
class Conversation
implements Runnable
{
   /** Logger */
   private static final Logger LOG =
      LogHelper.getLogger(Conversation.class.getName());
   
   /** An inbound message to process. */
   private Message msg = null;
   
   /** Request message if <code>true</code>, otherwise it is a reply. */
   private boolean request = false;
   
   /** The session-specific network transport. */
   private Queue queue = null;
   
   /** The singleton instance of the dispatcher. */
   private Dispatcher dispatcher = Dispatcher.getInstance();
   
   /** Security manager plugin for network resources */
   private NetResource nr = null;
   
   /**
    * Constructs an instance.
    * 
    * @param   queue
    *          Queue used by the conversation for message passing.
    */
   Conversation(Queue queue)
   {
      this.queue = queue;
      
      SessionManager sessMgr = SessionManager.get();
      sessMgr.setInitialContext();
      
      nr = (NetResource) sessMgr.getSecurityPlugin("net");
   }

   /**
    * Implementation of <code>Runnable</code> interface.  Initializes the
    * security context and then calls the main message processing loop for
    * this side of the conversation.
    */
   public void run()
   {
      SessionManager sessMgr = SessionManager.get();
      boolean inContext = false;
      boolean apContext = false;
      
      try
      {
         if (sessMgr.isRouter())
         {
            // establish the initial security context (this is only
            // needed here to allow the security cache to migrate each
            // thread to any newer initial context that may be set over time
            sessMgr.setInitialContext();
            inContext = true;
   
            // set up the security context which is valid for the life of this thread
            sessMgr.setContext(queue);
            apContext = true;
            
            // name the new thread so thread dumps are more meaningful
            String name = LogHelper.createThreadName("Conversation");
            Thread.currentThread().setName(name);
            
            // force better naming for the protocol threads (must be done here
            // from a thread that has the proper security context)
            queue.setProtocolNames();

            // allows listeners to perform any required tasks on the proper context AFTER it
            // has been created and before client code will be executed
            sessMgr.notifyInitialization(queue);
         }
         
         // clear traces of interrupted status, if any
         Thread.interrupted();
         
         block();
      }
      catch (ApplicationRequestedStop exc)
      {
         // normal termination of the queue
         if (LOG.isLoggable(Level.FINE))
         {
            String msg = (apContext) ? LogHelper.generate(exc.getMessage())
                                     : exc.getMessage();
                                     
            LOG.logp(Level.FINE, "Conversation.run()", "", msg);
         }
      }
      catch (Exception exc)
      {
         // log the error and exit the thread
         if (!(exc instanceof InterruptedException))
         {
            if (LOG.isLoggable(Level.WARNING))
            {
               LOG.log(Level.WARNING,
                       "--- exception in Conversation.run() ---",
                       exc);
            }
         }
      }

      finally
      {
         if (apContext)
         {
            sessMgr.restoreContext();
         }
         
         if (inContext)
         {
            // drop security context
            sessMgr.dropInitialContext();
         }
      }
   }
   
   /**
    * Halt the conversation.
    */
   synchronized void halt()
   {
      msg = null;
      notify();
   }
   
   /**
    * Notify the conversation that a request message has been received.
    *
    * @param    msg
    *           The incoming message.
    */
   synchronized void request(Message msg)
   {
      this.msg     = msg;
      this.request = true;
      notify();
   }
   
   /**
    * Notify the conversation that a reply message has been received.
    *
    * @param    msg
    *           The incoming message.
    */
   synchronized void reply(Message msg)
   {
      this.msg     = msg;
      this.request = false;
      notify();
   }
   
   /**
    * Block until a reply message is received where the request ID matches
    * the given ID.
    *
    * @param    reqId
    *           The request ID to wait for.
    *
    * @return   The reply message.
    */
   Message waitMessage(int reqId) 
   throws InterruptedException,
          RestrictedUseException,
          IllegalAccessException,
          NoSuchMethodException,
          ApplicationRequestedStop,
          ProtocolViolation
   {
      Message response = block();
      int replyId = response.getRequestID();
      
      if (replyId != reqId)
      {
         String err = "Expected ID " + reqId + " but rec'd ID " + replyId;
         throw new ProtocolViolation(err);
      }
      
      return response;
   }
   
   /**
    * This is the main message processing loop for this side of the
    * conversation.  The method is designed to be re-entrant to allow the
    * thread's stack to naturally handle the synchronous nature/state of 
    * transaction support.
    * <p>
    * This method will not return until it receives a reply message.
    *
    * @return   The reply ID of the message just received. 
    */
   private Message block()
   throws InterruptedException,
          RestrictedUseException,
          IllegalAccessException,
          NoSuchMethodException,
          ApplicationRequestedStop
   {
      Message msg = null;
      boolean request = false;
      
      while (queue != null && queue.isRunning())
      {
         synchronized (this)
         {
            // check if the reply has already arrived, if we wait in that case
            // then the notify will never occur because that has already been
            // processed
            if (this.msg == null)
            {
               wait();
            }
            
            // check for a halt
            if (this.msg == null)
            {
               throw new ApplicationRequestedStop();
            }
            
            msg = this.msg;
            request = this.request;
            
            // reset the state of the conversation to allow for blocking
            this.msg     = null;
            this.request = false;
         }
            
         if (request)
         {
            // this is an incoming request for service
            InboundRequest ir = new InboundRequest(msg, queue);
            dispatcher.processInbound(ir, false, nr);
         }
         else
         {
            // this must be a reply we are waiting for
            break;
         }
      }

      if (queue == null || !queue.isRunning())
         throw new ApplicationRequestedStop();

      return msg;
   }
}
