/*
** Module   : Authenticator.java
** Abstract : hooks to implement custom authentication logic
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description----------------------
** 001 NVS 20050208   @19696 Created the initial version.
** 002 NVS 20050914   @22709 Added clientFinalize() to the interface.
** 003 GES 20061003   @30107 Added support for return codes and cleaned
**                           up code formatting/javadoc.
** 004 GES 20061220   @31797 Removed use of AuthUIHelper since it is
**                           generally redundant with the idea of a
**                           client-side hook anyway.
** 005 SVL 20100923          Added input parameter "parameters" for the
**                           clientAuthHook function. 
** 006 OM  20131018          Added support for optionally returning a SessionListener.
*/

package com.goldencode.p2j.security;

import com.goldencode.p2j.net.SessionListener;
import java.util.Map;
import java.util.Set;

/**
 * This interface defines methods that provide a way to implement custom 
 * authentication logic.
 */
public interface Authenticator
extends SecurityConstants
{
   /**
    * Configures the Authenticator by setting the "option" parameter from directory.xml.
    *
    * @param   option
    *          The value of "option" entry for the auth plugin.
    */
   public void configure(String option);

   /**
    * Implements client side custom authentication logic.
    *
    * @param    parameters
    *           Additional configuration parameters.
    * @param    code
    *           The result of the most recent attempt to authenticate or
    *           <code>AUTH_RESULT_NONE</code> if this is the first attempt.
    *
    * @return   Array of bytes to be transmitted to the server as the
    *           authorization input. 
    */
   public byte[] clientAuthHook(Map<String, Object> parameters, int code);
   
   /**
    * Finalizes any resources allocated during authentication by the client.
    */
   public void clientFinalize();
   
   /**
    * Implements server side custom authentication logic.
    * <p>
    * Accepts the byte array produced by the client side authentication hook
    * as authentication input and custom parameters.
    *
    * @param    auth
    *           The authorization input from the client.
    * @param    entity
    *           Additional configuration parameters taken from the directory.
    *
    * @return   The result of the authentication processing.
    */
   public AuthenticationResponse serverAuthHook(byte[] auth, String entity);

   /**
    * Optionally the server-side may return a session listener. If some additional settings should
    * be done based on <code>Session</code> events (initialization / termination).
    * The initialization callback will be executed on the client thread, after the correct
    * context will be created.
    * The termination callback will be executed when the session ends (the client disconnects).
    *
    * @return  The current session listener (if there is any).
    */
   public SessionListener getSessionListener();

   /**
    * Returns a set of entities that this class handles.
    * They will later be processed one at a time.
    * This runs on the server-side of application.
    *
    * @return  The set of entities which need authentication that the Authenticator has detected
    *          by inspecting the directory or other means.
    *          <p>
    *          <code>null</code> if enumeration of entities could not be executed because of
    *          insufficient information or other non-fatal errors.
    */
   public Set<String> getAuthenticationEntities();
}
