/*
** Module   : CustomHookSample.java
** Abstract : authentication hook sample code
**
** Copyright (c) 2005-2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- -------------------------- Description ----------------------------
** 001 NVS 20050314   @20317 Created the file. This sample hook uses ID
**                           and password authentication as illustration.
** 002 NVS 20050901   @22445 Changed the decision making logic. Now it
**                           calls a method in SecurityManager.
** 003 NVS 20050914   @22707 Made it to comply with the changes to the
**                           Authenticator interface.
** 004 GES 20061003   @30108 More interface changes and cleanups.
** 005 GES 20061220   @31798 Matched interface changes.
** 006 SVL 20100923          Changed to conform the new Authenticator API.
** 007 OM  20131018          Updated to conform the new Authenticator API.
*/

package com.goldencode.p2j.security;

import java.io.*;
import java.util.*;
import com.goldencode.p2j.net.SessionListener;
import com.goldencode.p2j.util.*;

/**
 * This class is a sample of a custom authentication hook implementation. It
 * represents the minimum necessary to implement the interface.  Note that
 * no client-side user interface has been provided at all, that is an 
 * exercise left to the reader.
 */
class CustomHookSample
implements Authenticator
{
   /** The value of "option" node from AuthPlugin for this Authenticator */
   private String authOption = null;

   /**
    * Package private constructor. 
    */
   CustomHookSample()
   {
   }
   
   /**
    * Implements client side custom authentication logic. This simply uses
    * <code>stdin</code> and <code>stdout</code> to interact with the user.
    *
    * @param    parameters
    *           Additional configuration parameters.
    * @param    code
    *           The result of the most recent attempt to authenticate or
    *           <code>AUTH_RESULT_NONE</code> if this is the first attempt.
    *
    * @return   Array of bytes to be transmitted to the server as the
    *           authorization input. 
    */
   @Override
   public byte[] clientAuthHook(Map<String, Object> parameters, int code)
   {
      SecurityManager sm = SecurityManager.getInstance();
      
      String id = "";
      String pw = ""; 
      
      try
      {
         id = new String(Utils.prompt("User ID  :"));
         pw = new String(Utils.prompt("Password :"));
      }
      
      catch (IOException ioe)
      {
         // allow invalid input to pass through
      }
      
      return sm.packageIdPassword(id, pw);
   }
   
   /**
    * Implements server side custom authentication logic.
    * <p>
    * Accepts the byte array produced by the client side authentication hook
    * as authentication input as well as any custom parameters.
    *
    * @param    auth
    *           The authorization input from the client.
    * @param    entity
    *           Additional configuration parameters taken from the directory.
    *
    * @return   The result of the authentication processing.
    */
   @Override
   public AuthenticationResponse serverAuthHook(byte[] auth, String entity)
   {
      SecurityManager sm = SecurityManager.getInstance();
      return sm.serverAuthHook(auth, entity);
   }

   /**
    * Finalizes any resources allocated during authentication by the client.
    */
   @Override
   public void clientFinalize()
   {
   }

   /**
    * Always returns <code>null</code>.
    *
    * @return  Always <code>null</code>.
    */
   @Override
   public SessionListener getSessionListener()
   {
      return null;
   }

   /**
    * Returns a set of entities that this class handles. The list of entities / parameters will
    * be configured using other ways.
    *
    * @return  Always <code>null</code>.
    */
   @Override
   public Set<String> getAuthenticationEntities()
   {
      Set<String> ret = new HashSet<>();
      ret.add(authOption);
      return ret;
   }

   /**
    * Configures the Authenticator by setting the "option" parameter from directory.xml.
    *
    * @param   option
    *          The value of "option" entry for the auth plugin.
    */
   @Override
   public void configure(String option)
   {
      this.authOption = option;
   }
}
