/*
** Module   : Session.java
** Abstract : Maintains per connection data.
**
** Copyright (c) 2004-2009, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --JPRM-- ----------------Description-----------------
** 001 NVS 20050307   @20214 Created. The Session class maintains per 
**                           connection data. 
** 002 NVS 20050505   @21111 Added getCache() method that returns the ins-
**                           tance of the security cache this session is
**                           bound to. This helps in fixing the security
**                           cache refresh bug.
** 003 GES 20090610   @42763 Added timestamp of when the session was created
**                           which is essentially the same thing as when the
**                           account logon occurred.
** 004 GES 20090622   @42785 Removed unnecessary set and accessor.
*/ 

package com.goldencode.p2j.security;

/**
 * Package private class that maintains all per connection data. Particularly,
 * the reference to the generation of the <code>SecurityCache</code> that is
 * in effect when the session is created, is kept here. This allows for a safe
 * fading out of all older generations of <code>SecurityCache</code> as soon
 * as the last session referring to the old <code>SecurityCache</code> goes 
 * away.
 */
class Session
{
   /** An object that serves as an ID of this connection */
   private Object id = null;

   /** security context created for this session */
   private SecurityContext context = null;

   /** security cache in effect by the time of session creation */
   private SecurityCache cache = null;

   /** Timestamp of when this session was created. */
   private long timestamp = System.currentTimeMillis();
   
   /**
    * Package private constructor for client side Session.
    *
    * @param id
    *        session ID, normally the socket that maps the connection
    * @param cache
    *        reference to a <code>SecurityCache</code>
    */
   Session(Object id, SecurityCache cache)
   {
      this.id = id;
      this.cache = cache;
   }
   
   /**
    * Package private constructor for server side Session.
    *
    * @param id
    *        session ID, normally the socket that maps the connection
    * @param cache
    *        reference to a <code>SecurityCache</code>
    * @param context
    *        security context to be associated with the session
    */
   Session(Object id, SecurityCache cache, SecurityContext context)
   {
      this(id, cache);
      this.context = context;
   }
   
   /**
    * Checks whether this session was created with the specified security
    * context.
    *
    * @param  context
    *         security context to check
    * @return <code>true</code> if security context matches
    */
   boolean hasContext(SecurityContext context)
   {
      return this.context == context;
   }

   /**
    * Gets the session ID object.
    *
    * @return session ID object
    */
   Object getSessionId()
   {
      return id;
   }

   /**
    * Gets the session security context.
    *
    * @return session <code>SecurityContext</code>
    */
   SecurityContext getSecurityContext()
   {
      return context;
   }

   /**
    * Gets the security cache this session is bound to.
    *
    * @return session <code>SecurityCache</code>
    */
   SecurityCache getCache()
   {
      return cache;
   }
   
   /**
    * Obtain the timestamp of when this session was created.
    *
    * @return   Milliseconds since the January 1, 1970.
    */
   long getTimeStamp()
   {
      return timestamp;
   }
}
